@@ -14,6 +14,9 @@ from datasets.arrow_dataset import Dataset
 from datasets.dataset_dict import DatasetDict
 from datasets.load import load_dataset
 from tqdm.auto import tqdm
+from transformers.data.data_collator import DataCollator
+from transformers.modeling_utils import PreTrainedModel
+from transformers.tokenization_utils import PreTrainedTokenizer
 from transformers.trainer import Trainer
 from transformers.trainer_callback import (
     EarlyStoppingCallback,
@@ -23,14 +26,12 @@ from transformers.trainer_callback import (
 )
 from transformers.trainer_utils import IntervalStrategy
 from transformers.training_args import OptimizerNames, TrainingArguments
-from transformers.utils.import_utils import is_torch_tpu_available
 
 from .callbacks import NeverLeaveProgressCallback
 from .config import BenchmarkConfig, DatasetConfig, ModelConfig
 from .exceptions import InvalidBenchmark
 from .hf_hub import get_model_config
 from .model_loading import load_model
-from .protocols import DataCollator, Model, Tokenizer
 from .scores import log_scores
 from .types import SCORE_DICT
 from .utils import (
@@ -74,11 +75,13 @@ class BenchmarkDataset(ABC):
         self.dataset_config = dataset_config
         self.benchmark_config = benchmark_config
         self._metrics = {
-            metric_cfg.name: evaluate.load(metric_cfg.huggingface_id)
+            metric_cfg.name: evaluate.load(
+                metric_cfg.huggingface_id, cache_dir=self.benchmark_config.cache_dir
+            )
             for metric_cfg in dataset_config.task.metrics
         }
 
-    def benchmark(
+    def benchmark(  # noqa
         self,
         model_id: str,
     ) -> Tuple[SCORE_DICT, Dict[str, int]]:
@@ -127,59 +130,15 @@ class BenchmarkDataset(ABC):
             cache_dir=self.benchmark_config.cache_dir,
         )
 
-        # Store the number of parameters in the model, the maximum sequence length and
-        # the size of the model's vocabulary
-        num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
-        max_seq_length = tokenizer.model_max_length
-        if hasattr(model.config, "vocab_size"):
-            vocab_size = model.config.vocab_size
-        elif hasattr(tokenizer, "vocab_size"):
-            vocab_size = tokenizer.vocab_size
-        else:
-            vocab_size = -1
-
-        # Store the metadata in a dictionary
-        metadata_dict = dict(
-            num_model_parameters=num_params,
-            max_sequence_length=max_seq_length,
-            vocabulary_size=vocab_size,
-        )
-
-        # Log the metadata
-        logger.info(
-            f"The model has {num_params:,} parameters, a vocabulary size of "
-            f"{vocab_size:,} and a maximum sequence length of {max_seq_length:,}."
-        )
+        # Get the metadata
+        metadata_dict = self._get_metadata(model=model, tokenizer=tokenizer)
 
         # Load the data collator
         data_collator = self._load_data_collator(tokenizer)
 
-        # Load the data
-        train, val, test = self._load_data()
-
         # Set variable with number of iterations
         num_iter = 10 if not self.benchmark_config.testing else 2
 
-        # Get bootstrap sample indices
-        test_bidxs = rng.integers(0, len(test), size=(num_iter, len(test)))
-
-        # Get bootstrapped datasets
-        tests = [
-            Dataset.from_dict(test[test_bidxs[idx]])
-            for idx in range(test_bidxs.shape[0])
-        ]
-
-        # Preprocess the datasets
-        try:
-            params = dict(framework="pytorch", config=model.config, tokenizer=tokenizer)
-            prepared_train = self._preprocess_data(train, split="train", **params)
-            prepared_val = self._preprocess_data(val, split="val", **params)
-        except ValueError:
-            raise InvalidBenchmark("Preprocessing of the dataset could not be done.")
-
-        # Get the training arguments
-        training_args = self._get_training_args()
-
         # Set up progress bar
         itr = tqdm(
             iterable=range(num_iter),
@@ -187,6 +146,34 @@ class BenchmarkDataset(ABC):
             disable=not self.benchmark_config.progress_bar,
         )
 
+        # Load the data
+        train, val, test = self._load_data()
+
+        # Get bootstrap sample indices
+        test_bidxs = rng.integers(0, len(test), size=(num_iter, len(test)))
+
+        # Get bootstrapped datasets
+        tests = [test.select(test_bidxs[idx]) for idx in range(test_bidxs.shape[0])]
+
+        # Set up the preprocessing parameters
+        preprocess_params = dict(
+            framework="pytorch", config=model.config, tokenizer=tokenizer
+        )
+
+        # Prepare the train and validation datasets
+        try:
+            prepared_train = self._preprocess_data(
+                train, split="train", **preprocess_params
+            )
+            prepared_val = self._preprocess_data(val, split="val", **preprocess_params)
+        except ValueError:
+            raise InvalidBenchmark(
+                "Preprocessing of the training and validation datasets could not be "
+                "done."
+            )
+
+        bs: int = self.benchmark_config.batch_size
+        ga: int = 32 // bs
         scores: Dict[str, List[Dict[str, float]]] = defaultdict(list)
         for idx in itr:
 
@@ -208,11 +195,33 @@ class BenchmarkDataset(ABC):
 
             while True:
 
-                # Prepare the bootstrapped test dataset
+                # Get the boostrapped dataset for this iteration
                 test = tests[idx]
-                prepared_test = self._preprocess_data(test, split="test", **params)
+
+                # Prepare the test dataset
+                try:
+                    prepared_test = self._preprocess_data(
+                        test, split="test", **preprocess_params
+                    )
+                except ValueError:
+                    raise InvalidBenchmark(
+                        "Preprocessing of the test dataset could not be done."
+                    )
+
+                # Re-block terminal output, as it gets unblocked by the `transformers`
+                # package before training
+                block_terminal_output()
+
+                # Get the training arguments
+                training_args = self._get_training_args(iteration_idx=idx)
+
+                # Set the correct batch size and gradient accumulation
+                training_args.per_device_train_batch_size = bs
+                training_args.per_device_eval_batch_size = bs
+                training_args.gradient_accumulation_steps = ga
 
                 itr_scores = self._benchmark_single_iteration(
+                    iteration_idx=idx,
                     model_config=model_config,
                     train=train,
                     prepared_train=prepared_train,
@@ -234,16 +243,13 @@ class BenchmarkDataset(ABC):
                 else:
                     bs = training_args.per_device_train_batch_size
                     ga = training_args.gradient_accumulation_steps
+
                     bs, ga = handle_error(
                         e=itr_scores,
                         per_device_train_batch_size=bs,
                         gradient_accumulation_steps=ga,
                     )
 
-                    training_args.per_device_train_batch_size = bs
-                    training_args.per_device_eval_batch_size = bs
-                    training_args.gradient_accumulation_steps = ga
-
                     # Clear memory, to avoid memory issues
                     try:
                         del model
@@ -269,7 +275,37 @@ class BenchmarkDataset(ABC):
 
         return all_scores, metadata_dict
 
-    def _get_training_args(self) -> TrainingArguments:
+    def _get_metadata(
+        self, model: PreTrainedModel, tokenizer: PreTrainedTokenizer
+    ) -> Dict[str, int]:
+
+        # Store the number of parameters in the model, the maximum sequence length and
+        # the size of the model's vocabulary
+        num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
+        max_seq_length = tokenizer.model_max_length
+        if hasattr(model.config, "vocab_size"):
+            vocab_size = model.config.vocab_size
+        elif hasattr(tokenizer, "vocab_size"):
+            vocab_size = tokenizer.vocab_size
+        else:
+            vocab_size = -1
+
+        # Store the metadata in a dictionary
+        metadata_dict = dict(
+            num_model_parameters=num_params,
+            max_sequence_length=max_seq_length,
+            vocabulary_size=vocab_size,
+        )
+
+        # Log the metadata
+        logger.info(
+            f"The model has {num_params:,} parameters, a vocabulary size of "
+            f"{vocab_size:,} and a maximum sequence length of {max_seq_length:,}."
+        )
+
+        return metadata_dict
+
+    def _get_training_args(self, iteration_idx: int) -> TrainingArguments:
 
         # Set the logging strategy
         if self.benchmark_config.verbose:
@@ -285,6 +321,9 @@ class BenchmarkDataset(ABC):
         # Set batch size variable
         batch_size = 32 if not self.benchmark_config.testing else 1
 
+        # Set seed variable
+        seed = 4242 + iteration_idx
+
         # Initialise training arguments
         with warnings.catch_warnings():
             warnings.simplefilter("ignore", category=UserWarning)
@@ -297,7 +336,7 @@ class BenchmarkDataset(ABC):
                 logging_steps=30,
                 save_steps=30,
                 max_steps=10_000 if not self.benchmark_config.testing else 2,
-                report_to=None,
+                report_to="none",
                 save_total_limit=1,
                 per_device_train_batch_size=batch_size,
                 per_device_eval_batch_size=batch_size,
@@ -306,7 +345,7 @@ class BenchmarkDataset(ABC):
                 gradient_accumulation_steps=1,
                 load_best_model_at_end=True,
                 optim=OptimizerNames.ADAMW_TORCH,
-                seed=4242,
+                seed=seed,
                 no_cuda=self.benchmark_config.testing,
                 use_mps_device=use_mps_device,
                 fp16=False,
@@ -329,7 +368,7 @@ class BenchmarkDataset(ABC):
 
         # If the dataset turns out not to be a DatasetDict, then we raise an error
         if not isinstance(dataset_dict, DatasetDict):
-            raise ValueError(
+            raise InvalidBenchmark(
                 f"Expected `dataset_dict` to be a `DatasetDict`, but got "
                 f"{type(dataset_dict)}."
             )
@@ -365,6 +404,7 @@ class BenchmarkDataset(ABC):
 
     def _benchmark_single_iteration(
         self,
+        iteration_idx: int,
         model_config: ModelConfig,
         train: Dataset,
         prepared_train: Dataset,
@@ -373,12 +413,14 @@ class BenchmarkDataset(ABC):
         prepared_test: Dataset,
         data_collator: DataCollator,
         training_args: TrainingArguments,
-        tokenizer: Optional[Tokenizer] = None,
-        model: Optional[Model] = None,
+        tokenizer: Optional[PreTrainedTokenizer] = None,
+        model: Optional[PreTrainedModel] = None,
     ) -> Union[Dict[str, Dict[str, float]], Exception]:
         """Run a single iteration of a benchmark.
 
         Args:
+            iteration_idx (int):
+                The index of the iteration.
             model_config (ModelConfig):
                 The model configuration.
             train (Dataset):
@@ -395,10 +437,10 @@ class BenchmarkDataset(ABC):
                 The data collator.
             training_args (TrainingArguments):
                 The training arguments.
-            tokenizer (Tokenizer or None, optional):
+            tokenizer (PreTrainedTokenizer or None, optional):
                 The tokenizer to use in the benchmark. If None then a new tokenizer
                 will be loaded. Defaults to None.
-            model (Model or None, optional):
+            model (PreTrainedModel or None, optional):
                 The model to use in the benchmark. If None then a new model will be
                 loaded. Defaults to None.
 
@@ -412,7 +454,8 @@ class BenchmarkDataset(ABC):
         try:
             # Set random seeds to enforce reproducibility of the randomly initialised
             # weights
-            enforce_reproducibility(framework=model_config.framework)
+            seed = 4242 + iteration_idx
+            enforce_reproducibility(framework=model_config.framework, seed=seed)
 
             # Reinitialise a new model
             if tokenizer is None or model is None:
@@ -511,11 +554,11 @@ class BenchmarkDataset(ABC):
 
     def _get_trainer(
         self,
-        model: Model,
+        model: PreTrainedModel,
         args: TrainingArguments,
         train_dataset: Dataset,
         eval_dataset: Dataset,
-        tokenizer: Tokenizer,
+        tokenizer: PreTrainedTokenizer,
         data_collator: DataCollator,
         compute_metrics: Callable,
         callbacks: List[TrainerCallback],
@@ -523,7 +566,7 @@ class BenchmarkDataset(ABC):
         """Get a Trainer object.
 
         Args:
-            model (Model):
+            model (PreTrainedModel):
                 The model to finetune.
             args (TrainingArguments):
                 The training arguments.
@@ -531,7 +574,7 @@ class BenchmarkDataset(ABC):
                 The training dataset.
             eval_dataset (Dataset):
                 The evaluation dataset.
-            tokenizer (Tokenizer):
+            tokenizer (PreTrainedTokenizer):
                 The tokenizer.
             data_collator (DataCollator):
                 The data collator.
@@ -614,12 +657,12 @@ class BenchmarkDataset(ABC):
 
     @abstractmethod
     def _load_data_collator(
-        self, tokenizer: Optional[Tokenizer] = None
+        self, tokenizer: Optional[PreTrainedTokenizer] = None
     ) -> DataCollator:
         """Load the data collator used to prepare samples during finetuning.
 
         Args:
-            tokenizer (Tokenizer or None, optional):
+            tokenizer (PreTrainedTokenizer or None, optional):
                 A pretrained tokenizer. Can be None if the tokenizer is not used in the
                 initialisation of the data collator. Defaults to None.
 
